package extern;
import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;



/**
* Diese Klasse dient zur Definition triangulierter Bilder und zur Interpolation
* zwischen zwei passend triangulierten Bildern.
* 
* @author Frank Klawonn
* Letzte Aenderung 31.05.2005
* 
* @see MorphingCandS
*/
public class TriangulatedImage
{

  //Das Ausgangsbild
  BufferedImage bi;

  //Die Punkte, die fuer die Triangulation benoetigt werden. 
  Point2D[] tPoints;

  //Die Dreiecke, die die Triangulation definieren. Jede Zeile des Arrays enthaelt
  //die Indizes dreier Punkte des Arrays tPoints. Das i-te Dreieck der Triangulation
  //wird durch die drei Punkte des tPoints-Arrays mit den Indices
  //triangles[i][0], triangles[i][1], triangles[i][2]
  //definiert.
  int[][] triangles;



  /**
  * Eine Methode, die einen Punkt v und drei weitere nicht-kollineare Punkte 
  * erhaelt und v als Linearkombination der drei Punkte darstellt. Es werden
  * die drei Koeffizienten zur Darstellung von v zurueckgegeben.
  * 
  * @param v          darzustellender Punkt
  * @param triangle   ein Array, das aus drei nicht-kollinearen Punkten bestehen muss
  * 
  * @return           die drei Koeffizienten zur Darstellung von v
  */
  public static double[] triangleCoordinates(Point2D v, Point2D[] triangle)
  {
    //Das Array, das die drei zu berechnenden Koeffizienten enthalten wird.
  	double[] result = new double[3];

  	//Werte, die zur Loesung des Gleichungssystem zur Bestimmung der
  	//Koeffizienten erforderlich sind.
    double d13x = triangle[0].getX() - triangle[2].getX();
    double d23x = triangle[1].getX() - triangle[2].getX();
    double dx3 =  v.getX()           - triangle[2].getX();
    double d13y = triangle[0].getY() - triangle[2].getY();
    double d23y = triangle[1].getY() - triangle[2].getY();
    double dy3 =  v.getY()           - triangle[2].getY();

    double delta = d13x*d23y - d23x*d13y;

    //Wenn die drei Punkte (fast) kollinear sind, wird Division durch 0 vermieden und
    //Koeffizienten zurueckgegeben, die zumindest keine Konvexkombinaion repraesentieren.
    //Andernfalls koennen die Koeffizienten berechnet werden.
    if (Math.abs(delta)<0.00000001)
    {
      result[0] = 10;
    }
    else
    {
      result[0] = (dx3*d23y - d23x*dy3)/delta;
      result[1] = (d13x*dy3 - dx3*d13y)/delta;
    }


    result[2] = 1 - result[0] - result[1];

    return(result);
  }



  /**
  * Ueberpruefung, ob ein gegebenes Array von Koeffizienten eine Konvexkombination
  * darstellt, d.h., es wird ueberprueft, ob alle Array-Elemente zwischen 0 und 1 
  * liegen und ob alle nicht-negativ sind.
  * 
  * @param t          Array mit den Koeffizienten
  * 
  * @return           true genau dann, wenn die Koeffizienten eine Konvekombination
  *                   ergeben.
  */
  public static boolean isConvexCombination(double[] t)
  {
    boolean result;
    double sum;
    //Durch Rundungsfehler werden sich die Koeffizienten nicht genau zu 1 aufaddieren.
    //Der folgende Wert gibt an, wie gro?e Abweichungen von 1 noch toleriert werden.
    double tolerance = 0.000001;

    result = true;
    sum = 0;
    for (int i=0; i<t.length; i++)
    {
      if (t[i]<0 || t[i]>1) {result = false;}
      sum = sum + t[i];
    }
    
    if (Math.abs(sum-1)>tolerance) {result = false;}
    
    return(result);
  }



  
  /**
  * Gibt ein interpoliertes Zwischenbild aus zwei triangulierten Bildern (this und ti)
  * zurueck. ti hat einen Anteil von alpha*100% am Ergebnisbild. Die beiden Bilder 
  * muessen dieselbe Groesse, dieselbe Anzahl an Triangulationspunkten und Dreiecken
  * haben. Au?erdem muessen in beiden Bildern die Dreiecke aus den gleichen Tripeln
  * von Punkten zusammengesetzt sein. Nur die Lage der Punkte darf in dne beiden Bildern 
  * differieren.
  * 
  * @param ti         Zweites trianguliertes Bild
  * @param alpha      Anteil von ti am interpolierten Bild.
  * 
  * @return           interpoliertes Bild
  */
  public BufferedImage mixWith(TriangulatedImage ti, double alpha)
  {
    TriangulatedImage mix = new TriangulatedImage();

    //Dies wird das interpolierte Bild aufnehmen.
    mix.bi = new BufferedImage(this.bi.getWidth(),this.bi.getHeight(),
                                            BufferedImage.TYPE_INT_RGB);

    
    //Definition der Triangulationspunkte im interpolierten Bild als 
    //Konvexkombination der entsprechenden Punkte in den beiden zu 
    //interpolierenden Bildern.
    mix.tPoints = new Point2D[this.tPoints.length];
    for (int i=0; i<mix.tPoints.length; i++)
    {
      mix.tPoints[i] = new Point2D.Double((1-alpha)*this.tPoints[i].getX()
                                          + alpha*ti.tPoints[i].getX(),
                                          (1-alpha)*this.tPoints[i].getY()
                                          + alpha*ti.tPoints[i].getY());
    }

    //Die Dreiecke mussen aus den gleichen Punkten zusammengesetzt werden wie 
    //in den beiden anderen Bildern.
    mix.triangles = this.triangles;



    int rgbValueThis;
    int rgbValueTi;
    @SuppressWarnings("unused") int rgbValueMix;
    Color thisColour;
    Color tiColour;
    Color pixelColour;
    int rMix;
    int gMix;
    int bMix;
    int xInt;
    int yInt;
    double[] t = new double[3];
    double aux;
    Point2D[] threePoints = new Point2D[3];;
    Point2D.Double pixel = new Point2D.Double();
    int tNo;
    boolean notFound;

    
    //Bestimme fuer jedes Pixel des interpolierten Bildes die Frabe als
    //Konvexkombination der RGB-Werte der entsprechenden Pixel in den
    //beiden zu interpolierenden Bildern.
    for (int i=0; i<mix.bi.getWidth(); i++)
    {
      for (int j=0; j<mix.bi.getHeight(); j++)
      {
        //Fuer dieses Pixel muss die Farbe bestimmt werden.
        pixel.setLocation(i,j);

        //tNo wird verwendet, um den Index des Dreiecks zu bestimmen, in dem das
        //betrachtete Pixel liegt.
        tNo = 0;

        //Wenn das Dreieck gefunden wurde, wechselt notFound auf false.
        notFound = true;
        while(tNo<mix.triangles.length && notFound)
        {
        	
          //Bestimme die drei Punkte des Dreiecks Nr. tNo.
          for(int k=0; k<3; k++)
          {
            threePoints[k] = mix.tPoints[mix.triangles[tNo][k]];
          }

          //Bestimme die Koordinaten des Pixels bzgl. des Dreiecks tNo.
          t = triangleCoordinates(pixel,threePoints);

          //Pruefe, ob das Pixel innerhalb des Dreiecks liegt.
          if (isConvexCombination(t))
          {
            notFound = false;
          }
          else
          {
            tNo++;
          }
        }//endWhile
        //Die while-Schleife endet, wenn ein Index tNo eines Dreiecks gefunden
        //wurde, in dem das Pixel liegt, oder wenn nach dem Durchlauf aller 
        //Dreiecke kein entsprechendes Dreieck gefunden wurde.
        
        
        
        //Wenn ein passendes Dreieck gefunden wurde, kann die Farbe des Pixels
        //berechnet werden.
        if (!notFound)
        {
          //Bestimme die x-Koordinate des Pixels im ersten Bild (this) als 
          //Konvexkombination der Eckpunkte im entsprechenden Dreieck.
          aux = 0;
          for (int k=0; k<3; k++)
          {
            aux = aux + t[k]*this.tPoints[this.triangles[tNo][k]].getX();
          }
          xInt = (int) Math.round(aux);
          
          //Es muss sichergestellt werden, dass das Pixel durch Rundungsfehler
          //nicht ausserhalb des Bildes liegt.
          xInt = Math.max(0,xInt);
          xInt = Math.min(this.bi.getWidth()-1,xInt);


          //Alles entsprechend fuer die y-Koordinate.
          aux = 0;
          for (int k=0; k<3; k++)
          {
            aux = aux + t[k]*this.tPoints[this.triangles[tNo][k]].getY();
          }
          yInt = (int) Math.round(aux);
          yInt = Math.max(0,yInt);
          yInt = Math.min(this.bi.getHeight()-1,yInt);


          //Bestimme die Farbe des Pixels im ersten Bild.
          rgbValueThis = this.bi.getRGB(xInt,yInt);
          thisColour = new Color(rgbValueThis);



          //Fuehre alle Berechnungen entsprechend fuer das zweite Bild ti durch.
          aux = 0;
          for (int k=0; k<3; k++)
          {
            aux = aux + t[k]*ti.tPoints[ti.triangles[tNo][k]].getX();
          }
          xInt = (int) Math.round(aux);
          //Es muss sichergestellt werden, dass das Pixel durch Rundungsfehler
          //nicht ausserhalb des Bildes liegt.
          xInt = Math.max(0,xInt);
          xInt = Math.min(this.bi.getWidth()-1,xInt);

          aux = 0;
          for (int k=0; k<3; k++)
          {
            aux = aux + t[k]*ti.tPoints[ti.triangles[tNo][k]].getY();
          }
          yInt = (int) Math.round(aux);
          yInt = Math.max(0,yInt);
          yInt = Math.min(this.bi.getHeight()-1,yInt);
          
          rgbValueTi = ti.bi.getRGB(xInt,yInt);
          tiColour = new Color(rgbValueTi);



          //Now that we have the two colours, we can compute their
          //convex combimation for the R-,G- and B-value.
          //Round-off errors are quite unlikely here.
          //Nachdem die Farben der beiden entsprechenden Pixel in 
          //den zu interpolierenden Bildern bestimmt wurden, kann
          //die Konvexkombination der R-,G- und B-Werte berechnet werden.
          //Fehler durch Rundungsfehler sind durch das ganzzahlige Runden
          //sehr unwahrscheinlich.
          rMix = (int) Math.round((1-alpha)*thisColour.getRed()+
                                        alpha*tiColour.getRed());

          gMix = (int) Math.round((1-alpha)*thisColour.getGreen()+
                                        alpha*tiColour.getGreen());

          bMix = (int) Math.round((1-alpha)*thisColour.getBlue()+
                                        alpha*tiColour.getBlue());


          //Erzeuge die Mischfarbe als Konvexkombination.
          pixelColour = new Color(rMix, gMix, bMix);

          //Faerbe das Pixel im Interpolationsbild entsprechend.
          mix.bi.setRGB(i,j,pixelColour.getRGB());

        }//endif (!notFound)

      }//endfor j
    }//endfor i



    return(mix.bi);
  }



}

